=== modified file 'bzrlib/bundle/apply_bundle.py'
--- bzrlib/bundle/apply_bundle.py	2007-02-09 15:56:49 +0000
+++ bzrlib/bundle/apply_bundle.py	2007-06-11 14:59:52 +0000
@@ -25,6 +25,9 @@
 
 
 def install_bundle(repository, bundle_reader):
+    custom_install = getattr(bundle_reader, 'install', None)
+    if custom_install is not None:
+        return custom_install(repository)
     pb = bzrlib.ui.ui_factory.nested_progress_bar()
     repository.lock_write()
     try:

=== modified file 'bzrlib/bundle/serializer/v10.py'
--- bzrlib/bundle/serializer/v10.py	2007-06-11 12:10:43 +0000
+++ bzrlib/bundle/serializer/v10.py	2007-06-11 14:59:52 +0000
@@ -1,3 +1,4 @@
+from bzrlib import multiparent
 from bzrlib.bundle import serializer
 from bzrlib.util import bencode
 
@@ -6,19 +7,46 @@
     def write(self, repository, revision_ids, forced_bases, fileobj):
         container = _PseudoContainer(fileobj)
         transaction = repository.get_transaction()
-        for file_id, file_revision_ids in \
-            repository.fileids_altered_by_revision_ids(revision_ids).iteritems():
+        altered = repository.fileids_altered_by_revision_ids(revision_ids)
+        for file_id, file_revision_ids in altered.iteritems():
             vf = repository.weave_store.get_weave(file_id, transaction)
             file_revision_ids = [r for r in revision_ids if r in
                                  file_revision_ids]
             for file_revision_id in file_revision_ids:
                 text = ''.join(vf.make_mpdiff(file_revision_id).to_patch())
-                container.add_record('M', len(text), ['file:%s/%s' %
-                                     (file_id, file_revision_id)], text)
+                container_name = self.encode_name('file', file_revision_id,
+                                                  file_id)
+                container.add_record('M', len(text), [container_name], text)
         container.finish()
 
-    def read(self, file)
-        
+    def read(self, file):
+        container = _RecordReader(file, self)
+        return container
+
+    @staticmethod
+    def encode_name(name_kind, revision_id, file_id=None):
+        assert name_kind in ('revision', 'file')
+        if name_kind in ('revision',):
+            assert file_id is None
+        else:
+            assert file_id is not None
+        if file_id is not None:
+            file_tail = '/' + file_id
+        else:
+            file_tail = ''
+        return name_kind + ':' + revision_id + file_tail
+
+    @staticmethod
+    def decode_name(name):
+        kind, revisionfile_id = name.split(':', 1)
+        revisionfile_id = revisionfile_id.split('/')
+        if len(revisionfile_id) == 1:
+            revision_id = revisionfile_id[0]
+            file_id = None
+        else:
+            revision_id, file_id = revisionfile_id
+        return kind, revision_id, file_id
+
 
 class _PseudoContainer(object):
     
@@ -31,3 +59,57 @@
 
     def finish(self):
         self._fileobj.write(bencode.bencode([list(e) for e in self._records]))
+
+
+class _RecordReader(object):
+
+    def __init__(self, fileobj, serializer):
+        self._records = [tuple(e) for e in bencode.bdecode(fileobj.read())]
+        self._record_iter = iter(self._records)
+        self._current_text = None
+        self._serializer = serializer
+
+    def iter_records(self):
+        for type, size, names, text in self._records:
+            self._current_text = text
+            yield type, size, names
+        yield 'E', None, None
+
+    def read_record(self):
+        return self._current_text
+
+    def install(self, repository):
+        current_file = None
+        current_versionedfile = None
+        pending_file_records = []
+        for type_, size, names  in self.iter_records():
+            if type_ == 'E':
+                self._install_file_records(current_versionedfile,
+                    pending_file_records)
+                break
+            (name,) = names
+            kind, revision_id, file_id = self._serializer.decode_name(name)
+            if  kind != 'file':
+                self._install_file_records(current_versionedfile,
+                    pending_file_records)
+            if kind == 'file':
+                if file_id != current_file:
+                    self._install_file_records(current_versionedfile,
+                        pending_file_records)
+                    current_file = file_id
+                    current_versionedfile = \
+                        repository.weave_store.get_weave_or_empty(file_id,
+                        repository.get_transaction())
+                    pending_file_records = []
+                if revision_id in current_versionedfile:
+                    continue
+                pending_file_records.append((type_, revision_id, [],
+                                            self.read_record()))
+
+
+    def _install_file_records(self, current_versionedfile,
+                              pending_file_records):
+        for type_, revision, parents, text in pending_file_records:
+            assert type_ == 'M'
+            mpdiff = multiparent.MultiParent.from_lines(text)
+            current_versionedfile.add_mpdiff(revision, parents, mpdiff)

=== modified file 'bzrlib/multiparent.py'
--- bzrlib/multiparent.py	2007-06-11 02:35:08 +0000
+++ bzrlib/multiparent.py	2007-06-11 14:59:52 +0000
@@ -563,6 +563,7 @@
         length = self.diffs.get_diff(version_id).num_lines()
         return self._reconstruct(lines, version_id, 0, length)
 
+
 def gzip_string(lines):
     sio = StringIO()
     data_file = GzipFile(None, mode='wb', fileobj=sio)

=== modified file 'bzrlib/tests/test_bundle.py'
--- bzrlib/tests/test_bundle.py	2007-06-11 12:10:43 +0000
+++ bzrlib/tests/test_bundle.py	2007-06-11 14:59:52 +0000
@@ -1042,13 +1042,30 @@
     def test_creation(self):
         tree = self.make_branch_and_tree('tree')
         self.build_tree(['tree/file'])
-        tree.add('file')
+        tree.add('file', 'fileid-2')
         tree.commit('added file', rev_id='rev1')
         s = StringIO()
         serializer = BundleSerializerV10('1.0')
         serializer.write(tree.branch.repository, ['rev1'], {}, s)
         s.seek(0)
-        bundle_info = serializer.read(s)
+        tree2 = self.make_branch_and_tree('target')
+        target_repo = tree2.branch.repository
+        install_bundle(target_repo, serializer.read(s))
+        vf = target_repo.weave_store.get_weave('fileid-2',
+            target_repo.get_transaction())
+        self.assertEqual(tree.get_file_text('fileid-2'), vf.get_text('rev1'))
+
+    def test_name_encode(self):
+        self.assertEqual('revision:rev1',
+            BundleSerializerV10.encode_name('revision', 'rev1'))
+        self.assertEqual('file:rev1/file-id-1',
+            BundleSerializerV10.encode_name('file', 'rev1', 'file-id-1'))
+
+    def test_name_decode(self):
+        self.assertEqual(('revision', 'rev1', None),
+            BundleSerializerV10.decode_name('revision:rev1'))
+        self.assertEqual(('file', 'rev1', 'file-id-1'),
+            BundleSerializerV10.decode_name('file:rev1/file-id-1'))
 
 
 class MungedBundleTester(TestCaseWithTransport):

